#include "configreader.h"

#include "graphics/newci.h"
#include "graphics/newcrt.h"
#include "utils/strings.h"
#include "log.h"
#include "window.h"
#include "utils.h"
#include "tool.h"


static void removepath(char *name)
{
	char *cp = (char *)strrchr(name, '/');
	if (cp)
		strcpy_overlap(name, cp + 1);
}

void beginfatal(struct LOC_initialize *LINK)
{
	if (P_escapecode == -20)
		_Escape(P_escapecode);
	clearshowalpha();
	nc_printf("\n\n%c", chryellow);
}

void endfatal(struct LOC_initialize *LINK)
{
	nc_printf("%c  Program aborted.\n", chrgreen);
	_Escape(0);
}


void sethomedirname(char *dir_, struct LOC_initialize *LINK)
{
	char dir[256];
	char *ep;

	if (dir_[0] == '~')
	{
		ep = tilde_expand(dir_);
		strncpy(dir, ep, 256);
		Free(ep);
	}
	else
	{
		strncpy(dir, dir_, 256);
	}
	strcpy(gg.homedirname, dir);
	if (dir[strlen(dir) - 1] != ':' && dir[strlen(dir) - 1] != '/')
		strcat(gg.homedirname, "/");
}

void addgatesfile(char *fn_, struct LOC_initialize *LINK)
{
	char fn[256];
	strlist_t *l1;

	strcpy(fn, fn_);
	newci_fixfname(fn, "gate", "");
	if (!locatefile(fn, LINK))
		*fn = '\0';
	if (*fn != '\0' && strlist_find(gatefilenames, fn) == NULL)
		l1 = strlist_append(&gatefilenames, fn);
}

void readcnf(char *fn_, struct LOC_initialize *LINK)
{
	char fn[256];
	int flag;
	long cnflin;
	FILE *tx;
	char txbuf[256], txarg[256], txarg2[256];
	int erasegates, eraseload, erasemenu;
	strlist_t *l1, *l2;
	log_tool *tp;
	cnfrec *cnfp;
	char *TEMP;

	strcpy(fn, fn_);
	tx = NULL;
	erasegates = true;
	eraseload = true;
	erasemenu = true;
	flag = false;
	if (!strcmp(fn, "*"))
	{
		strcpy(fn, P_argv[0]);
		removepath(fn);
		strcat(fn, ".cnf");
		if (!locatefile(fn, LINK))
		{
			strcpy(fn, "log.cnf");
			if (!locatefile(fn, LINK))
				*fn = '\0';

		}
	}
	else
	{
		newci_fixfname(fn, "cnf", "");
		if (!locatefile(fn, LINK))
			*fn = '\0';
	}
	if (*fn != '\0')
	{
		TRY(try32);
		if (tx != NULL)
			tx = freopen(fn, "r", tx);
		else
			tx = fopen(fn, "r");

		if (tx == NULL) {
			P_escapecode = -10;
			P_ioresult = FileNotFound;
			goto _Ltry32;
		}
		flag = true;
		RECOVER2(try32,_Ltry32);
		if (P_escapecode == -20)
			_Escape(P_escapecode);

		ENDTRY(try32);
	}
	cnflin = 0;
	if (flag)
	{
		TRY(try33);
		while (!P_eof(tx))
		{
			cnflin++;
			fgets(txbuf, 256, tx);
			TEMP = (char *)strchr(txbuf, '\n');
			if (TEMP != NULL)
				*TEMP = 0;

			strcomment(txbuf, "{}");
			getword(txbuf, LINK->cmdbuf);
			if (!strcmp(LINK->cmdbuf, "BOBCAT"))
			{
				if (nc_gType() != nc_g300)
					*txbuf = '\0';

				getword(txbuf, LINK->cmdbuf);
			}
			else if (!strcmp(LINK->cmdbuf, "CHIPMUNK"))
			{
				if (nc_gType() == nc_g300)
					*txbuf = '\0';

				getword(txbuf, LINK->cmdbuf);
			}
			if (!strcmp(LINK->cmdbuf, "LOG:"))
				getword(txbuf, LINK->cmdbuf);

			strcpy(gg.func, LINK->cmdbuf);
			strcpy(gg.funcarg, txbuf);
			if (*LINK->cmdbuf != '\0')
			{
				doimmedcnffunction();
				docnffunction();
			}
			if (*LINK->cmdbuf == '\0' || strcmp(gg.func, LINK->cmdbuf))
				continue;
			if (strends(LINK->cmdbuf, ":") && strlen(LINK->cmdbuf) > 1)
			{
				if (*txbuf == '\0')
					continue;

				LINK->cmdbuf[strlen(LINK->cmdbuf) - 1] = '\0';
				cnfp = (cnfrec *)Malloc(sizeof(cnfrec));
				cnfp->next = NULL;
				if (LINK->cnflast == NULL)
					cnfbase = cnfp;
				else
					LINK->cnflast->next = cnfp;

				cnfp->next = NULL;
				strcpy(cnfp->tool, LINK->cmdbuf);
				strcpy(cnfp->s, txbuf);
				LINK->cnflast = cnfp;
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "INCLUDE"))
			{
				do
				{
					strwordx(txbuf, txarg);
					if (*txarg != '\0')
						readcnf(txarg, LINK);
				}
				while (*txarg != '\0');

				continue;
			}
			if (!strcmp(LINK->cmdbuf, "HOME"))
			{
				sethomedirname(txbuf, LINK);
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "GATES"))
			{
				strwordx(txbuf, txarg);
				if (!strcmp(txarg, "+"))
					strwordx(txbuf, txarg);
				else
				{
					if (erasegates)
						strlist_empty(&gatefilenames);
				}
				erasegates = false;
				while (*txarg != '\0')
				{
					addgatesfile(txarg, LINK);
					strwordx(txbuf, txarg);
				}
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "HELP"))
			{
				if (locatefile(txbuf, LINK))
					strcpy(loghelpname, txbuf);

				continue;
			}
			if (!strcmp(LINK->cmdbuf, "NEWS"))
			{
				if (locatefile(txbuf, LINK))
					strcpy(lognewsname, txbuf);

				continue;
			}
			if (!strcmp(LINK->cmdbuf, "TABLET"))
			{
				tabletaddr = getint(txbuf, tabletaddr);
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "MENU"))
			{
				getword(txbuf, txarg);
				if (!strcmp(txarg, "+"))
					getword(txbuf, txarg);
				else
				{
					if (erasemenu)
						strlist_empty(&LINK->logmenu);
				}
				erasemenu = false;
				while (*txarg != '\0')
				{
					l1 = strlist_append(&LINK->logmenu, txarg);
					getword(txbuf, txarg);
				}

				continue;
			}
			if (!strcmp(LINK->cmdbuf, "GET") || !strcmp(LINK->cmdbuf, "LOAD"))
			{
				getword(txbuf, txarg);
				if (!strcmp(txarg, "+"))
					getword(txbuf, txarg);
				else
				{
					if (eraseload)
						strlist_empty(&LINK->loadgates);
				}
				eraseload = false;
				while (*txarg != '\0')
				{
					l1 = strlist_append(&LINK->loadgates, txarg);
					getword(txbuf, txarg);
				}
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "GETGROUP"))
			{
				getword(txbuf, txarg);
				if (!strcmp(txarg, "+"))
					getword(txbuf, txarg);
				else
				{
					if (eraseload)
						strlist_empty(&LINK->loadgates);
				}
				eraseload = false;
				l1 = strlist_append(&LINK->loadgates, "");
				l1->value = (na_long)NULL;
				while (*txarg != '\0')
				{
					l2 = strlist_append((strlist_t **)(&l1->value), txarg);
					getword(txbuf, txarg);
				}
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "TOOL"))
			{
				getword(txbuf, txarg);
				if (*txarg == '\0')
					continue;

				tp = gg.toolbase;
				while (tp != NULL && strcmp(tp->name, txarg))
					tp = tp->next;

				if (tp != NULL)
					continue;

				newtool(&tp, txarg);
				strwordx(txbuf, txarg);
				if (*txarg != '\0')
					strcpy(tp->comment, txarg);

				getword(txbuf, txarg);
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "COMMAND"))
			{
				getword(txbuf, txarg2);
				if (*txarg2 == '\0')
					continue;

				do
				{
					getword(txbuf, txarg);
					if (*txarg != '\0')
					{
						l1 = strlist_add(&commandlist, txarg);
						*(char **)((char **)(&l1->value)) = strdup(txarg2);
					}
				}
				while (*txarg != '\0');

				continue;
			}
			if (!strcmp(LINK->cmdbuf, "SIGNALS") || !strcmp(LINK->cmdbuf, "NODES"))
			{
				LINK->j = getint(txbuf, 0L);
				if (LINK->j < 1)
				{
					P_escapecode = -1;
					goto _Ltry33;
				}
				gg.maxsignal = LINK->j;
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "DO"))
			{
				l2 = strlist_append(&thingstodo, gg.funcarg);
				continue;
			}
			if (!strcmp(LINK->cmdbuf, "UNDO"))
				strlist_empty(&thingstodo);
			else if (strcmp(LINK->cmdbuf, "CTOLERANCE") &&
					strcmp(LINK->cmdbuf, "CLOCK") &&
					strcmp(LINK->cmdbuf, "SCOPE") &&
					strcmp(LINK->cmdbuf, "AUTO") &&
					strcmp(LINK->cmdbuf, "POST") &&
					strcmp(LINK->cmdbuf, "PRINTER") &&
					strcmp(LINK->cmdbuf, "HELPIX") &&
					strcmp(LINK->cmdbuf, "GATESIX") &&
					strcmp(LINK->cmdbuf, "PLOTTER") &&
					strcmp(LINK->cmdbuf, "SIMMODE"))
			{
				P_escapecode = -1;
				goto _Ltry33;
			}
		}
		RECOVER2(try33,_Ltry33);
		if (P_escapecode == -20)
			_Escape(P_escapecode);

		if (P_escapecode != 0)
		{
			beginfatal(LINK);
			nc_printf("Error in line %ld of configuration file \"%s\"\n", cnflin, fn);
			endfatal(LINK);
		}
		ENDTRY(try33);
	}
	if (tx != NULL)
		fclose(tx);

	/*obsolete*/
}
