
#include <utils/p2c.h>
#include <graphics/newkbd.h>
#include <utils/sysdevs.h>
#include "logglobals.h"
#include "logdef.h"
#include "utils.h"
#include "wire.h"
#include "node.h"
#include "log.h"
#include "window.h"

void fryhwire(log_hwrec *hw)
{
	long t0;

	remcursor();
	clipon();
	t0 = sysclock();
	while (labs(sysclock() - t0) < frytime)
	{
		m_color((long)gg.color.conflict);
		hline(hw->x1, hw->x2, hw->y);
		m_color((long)gg.color.backgr);
		hline(hw->x1, hw->x2, hw->y);
	}
	clipoff();
	refreshsoon();
}


void fryvwire(log_vwrec *vw)
{
	long t0;

	remcursor();
	clipon();
	t0 = sysclock();
	while (labs(sysclock() - t0) < frytime)
	{
		m_color((long)gg.color.conflict);
		vline(vw->x, vw->y1, vw->y2);
		m_color((long)gg.color.backgr);
		vline(vw->x, vw->y1, vw->y2);
	}
	clipoff();
	refreshsoon();
}


/* Local variables for wantsolder: */
struct LOC_wantsolder
{
	int hasred, hasgreen, hasyellow, hasblue;
};

void markcolor(log_hwrec **hw, struct LOC_wantsolder *LINK)
{
	if (*hw == NULL)
		return;
	switch ((*hw)->wcolr)
	{

		case log_wcol_red:
			LINK->hasred = true;
			break;

		case log_wcol_green:
			LINK->hasgreen = true;
			break;

		case log_wcol_yellow:
			LINK->hasyellow = true;
			break;

		case log_wcol_blue:
			LINK->hasblue = true;
			break;
	}
}


/// Create a Horizontal Wire.
void newhw(log_hwrec **hw)
{
	*hw = (log_hwrec *)Malloc(sizeof(log_hwrec));
	(*hw)->next = gg.pages[gg.curpage - 1]->hwbase;
	gg.pages[gg.curpage - 1]->hwbase = *hw;
	(*hw)->temp = (na_long)0;
}


#define rtn             "DISPHW"


/// Dispose of a Horizontal Wire.
void disphw(log_hwrec **hw)
{
	log_hwrec *hw1;

	if (gg.traceflag)
		fprintf(tracefile, "Dispose hwire %ld\n", (long)(*hw));
	if (*hw == NULL)
		return;
	hw1 = gg.pages[gg.curpage - 1]->hwbase;
	if (*hw == hw1)
	{
		gg.pages[gg.curpage - 1]->hwbase = (*hw)->next;
	}
	else
	{
		while (hw1 != NULL && hw1->next != *hw)
			hw1 = hw1->next;
		if (hw1 == NULL)
			report(10, rtn);
		else
			hw1->next = (*hw)->next;
	}
	chpageplace(gg.pages[gg.curpage - 1], (*hw)->x1, (*hw)->y, (*hw)->x2, (*hw)->y);
	Free(*hw);
}

#undef rtn


/// Create a Vertical Wire.
void newvw(log_vwrec **vw)
{
	*vw = (log_vwrec *)Malloc(sizeof(log_vwrec));
	(*vw)->next = gg.pages[gg.curpage - 1]->vwbase;
	gg.pages[gg.curpage - 1]->vwbase = *vw;
	(*vw)->temp = (na_long)0;
}


#define rtn             "DISW"


/// Dispose of a Vertical Wire.
void dispvw(log_vwrec **vw)
{
	log_vwrec *vw1;

	if (gg.traceflag)
		fprintf(tracefile, "Dispose vwire %ld\n", (long)(*vw));
	if (*vw == NULL)
		return;
	vw1 = gg.pages[gg.curpage - 1]->vwbase;
	if (*vw == vw1)
	{
		gg.pages[gg.curpage - 1]->vwbase = (*vw)->next;
	}
	else
	{
		while (vw1 != NULL && vw1->next != *vw)
			vw1 = vw1->next;
		if (vw1 == NULL)
			report(10, rtn);
		else
			vw1->next = (*vw)->next;
	}
	chpageplace(gg.pages[gg.curpage - 1], (*vw)->x, (*vw)->y1, (*vw)->x, (*vw)->y2);
	Free(*vw);
}

#undef rtn

/// Create a solder point.
void newsolder(log_srec **s)
{
	*s = (log_srec *)Malloc(sizeof(log_srec));
	(*s)->next = gg.pages[gg.curpage - 1]->sbase;
	gg.pages[gg.curpage - 1]->sbase = *s;
	(*s)->hwire = NULL;
	(*s)->hwire2 = NULL;
	(*s)->vwire = NULL;
	(*s)->vwire2 = NULL;
	(*s)->temp = (na_long)0;
}


#define rtn             "DISPSOLDER"


/// Dispose of a solder point.
void dispsolder(log_srec **s)
{
	log_srec *s1;

	if (gg.traceflag)
		fprintf(tracefile, "Dispose solder %ld\n", (long)(*s));
	if (*s == NULL)
		return;
	s1 = gg.pages[gg.curpage - 1]->sbase;
	if (*s == s1)
	{
		gg.pages[gg.curpage - 1]->sbase = (*s)->next;
	}
	else
	{
		while (s1 != NULL && s1->next != *s)
			s1 = s1->next;
		if (s1 == NULL)
			report(10, rtn);
		else
			s1->next = (*s)->next;
	}
	chpageplace(gg.pages[gg.curpage - 1], (*s)->x, (*s)->y, (*s)->x, (*s)->y);
	Free(*s);
}

#undef rtn

/// Find which wire(s) are under the cursor, if any.
void closerwire(short x, short y)
{
	if (gg.invisible || gg.showconflicts)
	{
		gg.nearhw = NULL;
		gg.nearvw = NULL;
		return;
	}
	gg.nearhw = gg.pages[gg.curpage - 1]->hwbase;
	while (gg.nearhw != NULL &&
			(gg.nearhw->y != y || gg.nearhw->x1 > x || gg.nearhw->x2 < x))
		gg.nearhw = gg.nearhw->next;
	gg.nearvw = gg.pages[gg.curpage - 1]->vwbase;
	while (gg.nearvw != NULL &&
			(gg.nearvw->x != x || gg.nearvw->y1 > y || gg.nearvw->y2 < y))
		gg.nearvw = gg.nearvw->next;
}

/*assumes hw <> nil if hw2 <> nil*/

#define rtn             "ADDSOLDER"


/// Add a solder point between two wires.
void addsolder(short x, short y, log_hwrec *hw, log_hwrec *hw2, log_vwrec *vw, log_vwrec *vw2)
{
	log_srec *s;
	log_nrec *n;

	s = gg.pages[gg.curpage - 1]->sbase;
	while (s != NULL && (s->x != x || s->y != y))
		s = s->next;
	if (s == NULL) {
		newsolder(&s);
		s->x = x;
		s->y = y;
		chpageplace(gg.pages[gg.curpage - 1], x, y, x, y);
		s->hwire = NULL;
		s->hwire2 = NULL;
		s->vwire = NULL;
		s->vwire2 = NULL;
		n = NULL;
		if (hw != NULL)
			n = hw->node;
		if (vw != NULL && n == NULL)
			n = vw->node;
		hidecursor();
		if (gg.glowmode && n != NULL)
			drawsolderc(x, y, (int)glowcol(n, (long)gg.color.solder));
		else
			drawsolderc(x, y, gg.color.solder);
		hidecursor();
	}
	if (hw != NULL && hw != s->hwire && hw != s->hwire2)
	{
		if (s->hwire == NULL)
			s->hwire = hw;
		else if (s->hwire2 == NULL)
			s->hwire2 = hw2;
		else
			report(1, rtn);
	}
	if (hw2 != NULL && hw2 != s->hwire && hw2 != s->hwire2)
	{
		if (s->hwire2 == NULL)
			s->hwire2 = hw2;
		else
			report(2, rtn);
	}
	if (vw != NULL && vw != s->vwire && vw != s->vwire2)
	{
		if (s->vwire == NULL)
			s->vwire = vw;
		else if (s->vwire2 == NULL)
			s->vwire2 = vw;
		else
			report(3, rtn);
	}
	if (vw2 != NULL && vw2 != s->vwire && vw2 != s->vwire2)
	{
		if (s->vwire2 == NULL)
			s->vwire2 = vw2;
		else
			report(4, rtn);
	}
	if (gg.traceflag)
		fprintf(tracefile, "Add solder %ld\n", (long)s);
}

#undef rtn


/// Find a solder point at specified position.
log_srec *findsolder(short x, short y)
{
	log_srec *s;

	s = gg.pages[gg.curpage - 1]->sbase;
	while (s != NULL && (s->x != x || s->y != y))
		s = s->next;
	return s;
}


/// Add a solder point at a T-intersection of two wires.
void addsoldert(log_hwrec *hw, log_vwrec *vw)
{
	if (hw->x1 == vw->x && hw->y != vw->y1 && hw->y != vw->y2)
	{
		addsolder(hw->x1, hw->y, hw, NULL, vw, NULL);
		return;
	}
	if (hw->x2 == vw->x && hw->y != vw->y1 && hw->y != vw->y2)
	{
		addsolder(hw->x2, hw->y, hw, NULL, vw, NULL);
		return;
	}
	if (hw->y == vw->y1 && hw->x1 != vw->x && hw->x2 != vw->x)
		addsolder(vw->x, vw->y1, hw, NULL, vw, NULL);
	else if (hw->y == vw->y2 && hw->x1 != vw->x && hw->x2 != vw->x)
		addsolder(vw->x, vw->y2, hw, NULL, vw, NULL);
}


void chggate (log_grec *g, int i, log_nrec *oldnode, log_nrec *n);

void chgvw (log_vwrec *vw, log_nrec *oldnode, log_nrec *n);


#define rtn             "CHGHW"

/// Used for deleting wires (see DELHWIRE).
void chghw(log_hwrec *hw, log_nrec *oldnode, log_nrec *n)
{
	log_vwrec *vw;
	log_hwrec *hw1;
	log_grec *g;
	log_srec *s;
	short i, px, FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Change hwire %ld to node %ld\n", (long)hw, (long)n);
	switchnode(&hw->node, n);
	g = gg.pages[gg.curpage - 1]->gbase;
	while (g != NULL)
	{
		if (abs(g->y - hw->y) <= g->kind->bbmax &&
				P_ibetweenm((long)hw->x1, (long)g->x, (long)hw->x2,
					(long)g->kind->bbmax))
		{
			FORLIM = g->kind->numpins;
			for (i = 1; i <= FORLIM; i++)
			{
				px = g->pinpos[i - 1].x;
				if (hw->x1 <= px && px <= hw->x2 && g->pinpos[i - 1].y == hw->y &&
						g->pin[i - 1] == oldnode)
					chggate(g, i, oldnode, n);
			}
		}
		g = g->next;
	}
	vw = gg.pages[gg.curpage - 1]->vwbase;
	while (vw != NULL)
	{
		if (hw->x1 <= vw->x && vw->x <= hw->x2 && vw->y1 <= hw->y &&
				hw->y <= vw->y2 && vw->node == oldnode)
		{
			if ((hw->x1 == vw->x || hw->x2 == vw->x || vw->y1 == hw->y ||
						vw->y2 == hw->y) &&
					(hw->wcolr == vw->wcolr || hw->wcolr == log_wcol_blue ||
					 vw->wcolr == log_wcol_blue))
				chgvw(vw, oldnode, n);
			else
			{  /*vlsi only*/
				s = gg.pages[gg.curpage - 1]->sbase;
				while (s != NULL && (s->x != vw->x || s->y != hw->y))
					s = s->next;
				if (s != NULL)
					chgvw(vw, oldnode, n);
			}
		}
		vw = vw->next;
	}
	hw1 = gg.pages[gg.curpage - 1]->hwbase;
	while (hw1 != NULL && vlsi)
	{
		if (hw->y == hw1->y && (hw->x1 == hw1->x2 || hw->x2 == hw1->x1) &&
				(hw->wcolr == log_wcol_blue || hw1->wcolr == log_wcol_blue) &&
				hw1->node == oldnode)
			chghw(hw1, oldnode, n);
		hw1 = hw1->next;
	}
}

#undef rtn


#define rtn             "CHGVW"

// Used for deleting wires (see DELHWIRE).
void chgvw(log_vwrec *vw, log_nrec *oldnode, log_nrec *n)
{
	log_hwrec *hw;
	log_vwrec *vw1;
	log_grec *g;
	log_srec *s;
	short i, py, FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Change vwire %ld to node %ld\n", (long)vw, (long)n);
	switchnode(&vw->node, n);
	g = gg.pages[gg.curpage - 1]->gbase;
	while (g != NULL)
	{
		if (abs(g->x - vw->x) <= g->kind->bbmax &&
				P_ibetweenm((long)vw->y1, (long)g->y, (long)vw->y2,
					(long)g->kind->bbmax))
		{
			FORLIM = g->kind->numpins;
			for (i = 1; i <= FORLIM; i++)
			{
				py = g->pinpos[i - 1].y;
				if (vw->y1 <= py && py <= vw->y2 && g->pinpos[i - 1].x == vw->x &&
						g->pin[i - 1] == oldnode)
					chggate(g, i, oldnode, n);
			}
		}
		g = g->next;
	}
	hw = gg.pages[gg.curpage - 1]->hwbase;
	while (hw != NULL)
	{
		if (hw->x1 <= vw->x && vw->x <= hw->x2 && vw->y1 <= hw->y &&
				hw->y <= vw->y2 && hw->node == oldnode) {
			if (hw->x1 == vw->x || hw->x2 == vw->x || vw->y1 == hw->y ||
					vw->y2 == hw->y)
				chghw(hw, oldnode, n);
			else
			{
				s = gg.pages[gg.curpage - 1]->sbase;
				while (s != NULL && (s->x != vw->x || s->y != hw->y))
					s = s->next;
				if (s != NULL)
					chghw(hw, oldnode, n);
			}
		}
		hw = hw->next;
	}
	vw1 = gg.pages[gg.curpage - 1]->vwbase;
	while (vw1 != NULL && vlsi)
	{
		if (vw->x == vw1->x && (vw->y1 == vw1->y2 || vw->y2 == vw1->y1) &&
				(vw->wcolr == log_wcol_blue || vw1->wcolr == log_wcol_blue) &&
				vw1->node == oldnode)
			chgvw(vw1, oldnode, n);
		vw1 = vw1->next;
	}
}

#undef rtn

int wantsolder(log_srec *s)
{
	struct LOC_wantsolder V;

	V.hasred = false;
	V.hasgreen = false;
	V.hasyellow = false;
	V.hasblue = false;
	markcolor(&s->hwire, &V);
	markcolor(&s->hwire2, &V);
	markcolor((log_hwrec **)(&s->vwire), &V);
	markcolor((log_hwrec **)(&s->vwire2), &V);
	return (V.hasblue && (V.hasgreen || V.hasyellow || V.hasred));
}

/// Deletes a horizontal wire.
void delhwire(log_hwrec *hw)
{
	log_vwrec *vw;
	log_hwrec *hw1;
	log_srec *s, *s1;
	log_nrec *oldnode, *n, *savenode;
	log_grec *g;
	short i, x, y, FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Delete hwire %ld\n", (long)hw);
	clipon();
	m_color((long)gg.color.backgr);
	hline(hw->x1, hw->x2, hw->y);
	oldnode = hw->node;
	stamp(&oldnode->simtype->netstamp);
	switchnode(&hw->node, NULL);
	savenode = oldnode;
	s = gg.pages[gg.curpage - 1]->sbase;
	while (s != NULL)
	{
		s1 = s->next;
		if (s->hwire == hw || s->hwire2 == hw)
		{
			if (s->hwire == hw)
				s->hwire = s->hwire2;
			s->hwire2 = NULL;
			if (!wantsolder(s))
			{
				drawsolderc(s->x, s->y, gg.color.backgr);
				dispsolder(&s);
			}
			else
			{
				if (gg.traceflag)
					fprintf(tracefile, "DELHW keeping solder %ld\n", (long)s);
			}
		}
		s = s1;
	}
	vw = gg.pages[gg.curpage - 1]->vwbase;
	while (vw != NULL)
	{
		if (hw->x1 <= vw->x && vw->x <= hw->x2 && vw->y1 <= hw->y &&
				hw->y <= vw->y2 && vw->node == oldnode)
		{
			delgetnode(&savenode, &oldnode, &n);
			if (n != oldnode)
				chgvw(vw, oldnode, n);
		}
		vw = vw->next;
	}
	hw1 = gg.pages[gg.curpage - 1]->hwbase;
	while (hw1 != NULL && vlsi)
	{
		if (hw->y == hw1->y && (hw1->x1 == hw->x2 || hw1->x2 == hw->x1) &&
				hw1->node == oldnode) {
			delgetnode(&savenode, &oldnode, &n);
			if (n != oldnode)
				chghw(hw1, oldnode, n);
		}
		hw1 = hw1->next;
	}
	g = gg.pages[gg.curpage - 1]->gbase;
	while (g != NULL)
	{
		if (abs(g->y - hw->y) <= g->kind->bbmax &&
				P_ibetweenm((long)hw->x1, (long)g->x, (long)hw->x2,
					(long)g->kind->bbmax))
		{
			FORLIM = g->kind->numpins;
			for (i = 1; i <= FORLIM; i++)
			{
				if (g->pin[i - 1] == oldnode)
				{
					x = g->pinpos[i - 1].x;
					y = g->pinpos[i - 1].y;
					if (hw->x1 <= x && x <= hw->x2 && y == hw->y)
					{
						delgetnode(&savenode, &oldnode, &n);
						if (n != oldnode)
							chggate(g, i, oldnode, n);
					}
				}
			}
		}
		g = g->next;
	}
	disphw(&hw);
	clipoff();
}

/// Deletes a vertical wire.
void delvwire(log_vwrec *vw)
{
	log_hwrec *hw;
	log_vwrec *vw1;
	log_srec *s, *s1;
	log_nrec *oldnode, *n, *savenode;
	log_grec *g;
	short i, x, y, FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Delete vwire %ld\n", (long)vw);
	clipon();
	m_color((long)gg.color.backgr);
	vline(vw->x, vw->y1, vw->y2);
	oldnode = vw->node;
	stamp(&oldnode->simtype->netstamp);
	switchnode(&vw->node, NULL);
	savenode = oldnode;
	s = gg.pages[gg.curpage - 1]->sbase;
	while (s != NULL)
	{
		s1 = s->next;
		if (s->vwire == vw || s->vwire2 == vw)
		{
			if (s->vwire == vw)
				s->vwire = s->vwire2;
			s->vwire2 = NULL;
			if (!wantsolder(s)) {
				drawsolderc(s->x, s->y, gg.color.backgr);
				dispsolder(&s);
			}
			else
			{
				if (gg.traceflag)
					fprintf(tracefile, "DELVW keeping solder %ld\n", (long)s);
			}
		}
		s = s1;
	}
	hw = gg.pages[gg.curpage - 1]->hwbase;
	while (hw != NULL)
	{
		if (hw->x1 <= vw->x && vw->x <= hw->x2 && vw->y1 <= hw->y &&
				hw->y <= vw->y2 && hw->node == oldnode)
		{
			delgetnode(&savenode, &oldnode, &n);
			if (n != oldnode)
				chghw(hw, oldnode, n);
		}
		hw = hw->next;
	}
	vw1 = gg.pages[gg.curpage - 1]->vwbase;
	while (vw1 != NULL && vlsi)
	{
		if (vw->x == vw1->x && (vw1->y1 == vw->y2 || vw1->y2 == vw->y1) &&
				vw1->node == oldnode)
		{
			delgetnode(&savenode, &oldnode, &n);
			if (n != oldnode)
				chgvw(vw1, oldnode, n);
		}
		vw1 = vw1->next;
	}
	g = gg.pages[gg.curpage - 1]->gbase;
	while (g != NULL)
	{
		if (abs(g->x - vw->x) <= g->kind->bbmax &&
				P_ibetweenm((long)vw->y1, (long)g->y, (long)vw->y2,
					(long)g->kind->bbmax))
		{
			FORLIM = g->kind->numpins;
			for (i = 1; i <= FORLIM; i++)
			{
				if (g->pin[i - 1] == oldnode)
				{
					x = g->pinpos[i - 1].x;
					y = g->pinpos[i - 1].y;
					if (vw->y1 <= y && y <= vw->y2 && x == vw->x)
					{
						delgetnode(&savenode, &oldnode, &n);
						if (n != oldnode)
							chggate(g, i, oldnode, n);
					}
				}
			}
		}
		g = g->next;
	}
	vw1 = vw;
	dispvw(&vw);
	clipoff();
}


#define rtn             "ADDHWIRE"

/** @brief Add a Horizontal wire.

  Merge with other horizontal wires if touching;
  solder to other vertical wires in T-intersections;
  connect to gate pins as necessary.
  VLSI-mode rules are more complex!
  */
void addhwire(short x1, short x2, short y, short colr)
{
	cnrec *cnbase;
	log_hwrec *hw, *hw1, *hw2;
	log_vwrec *vw1, *vw2;
	log_grec *g;
	short i, xp, hx1, hx2, hy, vx, vy1, vy2, hcolr, vcolr;
	blobrec *blbase;
	short FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Add hwire %d-%d, %d\n", x1, x2, y);
	sortshints(&x1, &x2);
	cnbase = NULL;
	remcursor();
	clipon();
	if (!vlsi)
		colr = log_wcol_normal;
	newhw(&hw);
	hw->x1 = log_maxshint;
	hw->x2 = log_maxshint;
	hw->y = log_maxshint;
	hw->node = NULL;
	hw1 = hw->next;
	blbase = NULL;
	while (hw1 != NULL)
	{
		hw2 = hw1->next;
		hx1 = hw1->x1;
		hx2 = hw1->x2;
		hy = hw1->y;
		hcolr = hw1->wcolr;
		if (hy == y && hx1 <= x2 && hx2 >= x1)
		{
			if (hcolr == colr)
			{
				x1 = P_imin2((long)x1, (long)hx1);
				x2 = P_imax2((long)x2, (long)hx2);
				addblobs(&blbase, hx1, hy, hx2, hy);
				delhwire(hw1);
			}
		}
		hw1 = hw2;
	}
	hw->x1 = x1;
	hw->x2 = x2;
	hw->y = y;
	hw->wcolr = colr;
	chpageplace(gg.pages[gg.curpage - 1], x1, y, x2, y);
	g = gg.pages[gg.curpage - 1]->gbase;
	while (g != NULL)
	{
		if (abs(g->y - y) <= g->kind->bbmax &&
				P_ibetweenm((long)x1, (long)g->x, (long)x2, (long)g->kind->bbmax))
		{
			FORLIM = g->kind->numpins;
			for (i = 0; i < FORLIM; i++)
			{
				xp = g->pinpos[i].x;
				if (g->pinpos[i].y == y && x1 <= xp && xp <= x2)
					queuecombine(&cnbase, &g->pin[i], &hw->node);
			}
		}
		g = g->next;
	}
	vw1 = gg.pages[gg.curpage - 1]->vwbase;
	while (vw1 != NULL)
	{
		vx = vw1->x;
		vy1 = vw1->y1;
		vy2 = vw1->y2;
		vcolr = vw1->wcolr;
		vw2 = vw1->next;
		if ((((vx == x1 || vx == x2) && vy1 <= y && y <= vy2) ||
					((vy1 == y || vy2 == y) && x1 <= vx && vx <= x2)) &&
				(colr == vcolr || colr == log_wcol_blue || vcolr == log_wcol_blue))
			queuecombine(&cnbase, &vw1->node, &hw->node);
		else if (vlsi && x1 < vx && vx < x2 && vy1 < y && y < vy2 && colr == vcolr)
			queuecombine(&cnbase, &vw1->node, &hw->node);
		vw1 = vw2;
	}
	if (cnbase == NULL)
		newnoderef(&hw->node, 0, 1L);
	if (!checkcombine(&cnbase))
	{
		fryhwire(hw);
		disphw(&hw);
		gg.nearhw = NULL;
		goto _L1;   /*return*/
	}
	vw1 = gg.pages[gg.curpage - 1]->vwbase;
	while (vw1 != NULL)
	{
		vx = vw1->x;
		vy1 = vw1->y1;
		vy2 = vw1->y2;
		vcolr = vw1->wcolr;
		vw2 = vw1->next;
		if ((((vx == x1 || vx == x2) && vy1 <= y && y <= vy2) ||
					((vy1 == y || vy2 == y) && x1 <= vx && vx <= x2)) &&
				(colr == vcolr || colr == log_wcol_blue || vcolr == log_wcol_blue))
		{
			clipon();
			if (!vlsi)
				addsoldert(hw, vw1);
			else if (colr != vcolr)
				addsolder(vx, y, hw, NULL, vw1, NULL);
		}
		vw1 = vw2;
	}
	clipon();
	if (hw->node->conflict1 && hw->node->conflict2)   /*or hw^.node^.oconflict*/
		m_color((long)gg.color.conflict);
	else if (gg.glowmode)
		m_color(glowcol(hw->node, (long)gg.color.wire[colr - log_wcol_normal]));
	else
		m_color((long)gg.color.wire[colr - log_wcol_normal]);
	hline(x1, x2, y);
	doblobs(blbase);
	gg.nearhw = hw;
	clipoff();
	refreshsoon();
	if (gg.traceflag)
		fprintf(tracefile, "Added hwire %ld\n", (long)hw);
_L1:
	dispblobs(&blbase);

	/* Deal with color A drawn over color B */
	/*VVV*/
}

#undef rtn


void addhwire2(short x1, short x2, short y)
{
	addhwire(x1, x2, y, log_wcol_normal);
}


#define rtn             "ADDHWIRE"

/** @brief Add a vertical wire.

  Merge with other vertical wires if touching;
  solder to other horizontal wires in T-intersections;
  connect to gate pins as necessary.
  VLSI-mode rules are more complex!
  */
void addvwire(short x, short y1, short y2, short colr)
{
	cnrec *cnbase;
	log_vwrec *vw, *vw1, *vw2;
	log_hwrec *hw1, *hw2;
	log_grec *g;
	short i, yp, hy, hx1, hx2, vx, vy1, vy2, hcolr, vcolr;
	blobrec *blbase;
	short FORLIM;

	if (gg.traceflag)
		fprintf(tracefile, "Add vwire %d, %d-%d\n", x, y1, y2);
	sortshints(&y1, &y2);
	cnbase = NULL;
	remcursor();
	clipon();
	if (!vlsi)
		colr = log_wcol_normal;
	newvw(&vw);
	vw->x = log_maxshint;
	vw->y1 = log_maxshint;
	vw->y2 = log_maxshint;
	vw->node = NULL;
	vw1 = vw->next;
	blbase = NULL;
	while (vw1 != NULL)
	{
		vw2 = vw1->next;
		vx = vw1->x;
		vy1 = vw1->y1;
		vy2 = vw1->y2;
		vcolr = vw1->wcolr;
		if (vx == x && vy1 <= y2 && vy2 >= y1)
		{
			if (vcolr == colr)
			{
				y1 = P_imin2((long)y1, (long)vy1);
				y2 = P_imax2((long)y2, (long)vy2);
				addblobs(&blbase, vx, vy1, vx, vy2);
				delvwire(vw1);
			}
		}
		vw1 = vw2;
	}
	vw->y1 = y1;
	vw->y2 = y2;
	vw->x = x;
	vw->wcolr = colr;
	chpageplace(gg.pages[gg.curpage - 1], x, y1, x, y2);
	g = gg.pages[gg.curpage - 1]->gbase;
	while (g != NULL)
	{
		if (abs(g->x - x) <= g->kind->bbmax &&
				P_ibetweenm((long)y1, (long)g->y, (long)y2, (long)g->kind->bbmax))
		{
			FORLIM = g->kind->numpins;
			for (i = 0; i < FORLIM; i++)
			{
				yp = g->pinpos[i].y;
				if (g->pinpos[i].x == x && y1 <= yp && yp <= y2)
					queuecombine(&cnbase, &g->pin[i], &vw->node);
			}
		}
		g = g->next;
	}
	hw1 = gg.pages[gg.curpage - 1]->hwbase;
	while (hw1 != NULL)
	{
		hy = hw1->y;
		hx1 = hw1->x1;
		hx2 = hw1->x2;
		hcolr = hw1->wcolr;
		hw2 = hw1->next;
		if ((((hy == y1 || hy == y2) && hx1 <= x && x <= hx2) ||
					((hx1 == x || hx2 == x) && y1 <= hy && hy <= y2)) &&
				(colr == hcolr || colr == log_wcol_blue || hcolr == log_wcol_blue))
			queuecombine(&cnbase, &hw1->node, &vw->node);
		else if (vlsi && y1 < hy && hy < y2 && hx1 < x && x < hx2 && colr == hcolr)
			queuecombine(&cnbase, &hw1->node, &vw->node);
		hw1 = hw2;
	}

	if (cnbase == NULL)
		newnoderef(&vw->node, 0, 1L);
	if (!checkcombine(&cnbase))
	{
		fryvwire(vw);
		dispvw(&vw);
		gg.nearvw = NULL;
		goto _L1;   /*return*/
	}
	hw1 = gg.pages[gg.curpage - 1]->hwbase;
	while (hw1 != NULL)
	{
		hy = hw1->y;
		hx1 = hw1->x1;
		hx2 = hw1->x2;
		hcolr = hw1->wcolr;
		hw2 = hw1->next;
		if ((((hy == y1 || hy == y2) && hx1 <= x && x <= hx2) ||
					((hx1 == x || hx2 == x) && y1 <= hy && hy <= y2)) &&
				(colr == hcolr || colr == log_wcol_blue || hcolr == log_wcol_blue))
		{
			clipon();
			if (!vlsi)
				addsoldert(hw1, vw);
			else if (colr != hcolr)
				addsolder(x, hy, hw1, NULL, vw, NULL);
		}
		hw1 = hw2;
	}
	clipon();
	if (vw->node->conflict1 && vw->node->conflict2)   /*or vw^.node^.oconflict*/
		m_color((long)gg.color.conflict);
	else if (gg.glowmode)
		m_color(glowcol(vw->node, (long)gg.color.wire[colr - log_wcol_normal]));
	else
		m_color((long)gg.color.wire[colr - log_wcol_normal]);
	vline(x, y1, y2);
	doblobs(blbase);
	gg.nearvw = vw;
	clipoff();
	refreshsoon();
	if (gg.traceflag)
		fprintf(tracefile, "Added vwire %ld\n", (long)vw);
_L1:
	dispblobs(&blbase);

	/* Deal with color A drawn over color B */
	/*VVV*/
}

#undef rtn


void addvwire2(short x, short y1, short y2)
{
	addvwire(x, y1, y2, log_wcol_normal);
}

void unsoldernear()
{
	short oldx, oldy1, oldy2, oldcolr;
	blobrec *blbase;

	oldx = gg.nearvw->x;
	oldy1 = gg.nearvw->y1;
	oldy2 = gg.nearvw->y2;
	oldcolr = gg.nearvw->wcolr;
	if (gg.nearhw->y == oldy1 || gg.nearhw->y == oldy2 ||
			gg.nearhw->x1 == oldx || gg.nearhw->x2 == oldx)
		return;
	remcursor();
	blbase = NULL;
	addblobs(&blbase, oldx, oldy1, oldx, gg.nearhw->y - 1);
	addblobs(&blbase, oldx, gg.nearhw->y + 1, oldx, oldy2);
	delvwire(gg.nearvw);
	addvwire(oldx, oldy1, oldy2, oldcolr);
	doblobs(blbase);
	dispblobs(&blbase);
}


void unsolderwires(log_hwrec *hw, log_vwrec *vw)
{
	if (hw == NULL || vw == NULL)
		return;
	gg.nearhw = hw;
	gg.nearvw = vw;
	unsoldernear();
}


void soldernear()
{
	if (!trycombinenodes(&gg.nearvw->node, &gg.nearhw->node))
	{
		frysolder(gg.nearvw->x, gg.nearhw->y);
		return;
	}
	clipon();
	addsolder(gg.nearvw->x, gg.nearhw->y, gg.nearhw, NULL, gg.nearvw, NULL);
	clipoff();
}


void solderat(short x, short y)
{
	if (findsolder(x, y) != NULL)
		return;
	closerwire(x, y);
	if (gg.nearhw != NULL && gg.nearvw != NULL && gg.nearhw->x1 != x &&
			gg.nearhw->x2 != x && gg.nearvw->y1 != y && gg.nearvw->y2 != y)
		soldernear();
}

